# -*- coding: utf-8 -*-
#!/usr/bin/env python

"""
This module has two important functions:

htmlparser.find_download_links(): this function is used to scrape webpages for download links and html tables,
                                  allowing users to find and download items from a page.

htmlparser.compare(): this function is used to compare some data frame with a date column in it to the years in office
                      of the relevant president or party in office at the time.
"""

# \/ Local Imports \/
from .tablereader import get_fe

# \/ Third-Party Imports \/
from bokeh import plotting
import bokeh.palettes as palettes
from bokeh.embed import components
from flask import render_template
from urllib.parse import urlparse
from bs4 import BeautifulSoup
import pandas as pd
import os
import urllib.request
import logging
import datetime
from dateparser import parse

logging.basicConfig(filename='webservice.log', level=logging.DEBUG )  # set up the logger at level debug


def prlog(msg, log=True, prnt=False, level='ERROR'):
    """logging/printing function
    :type log: bool
    :param msg: msg to print/log
    :param log: if the function should log the msg or not
    :param prnt: if the function should print the msg or not
    :param level: what level the logger records the message at
    :return: None
    """

    if prnt:
        print(msg)
    elif log:
        print("logging." + level)
        logging.error(msg)



def download_approved(ext, filep):
    """checks if the file extension is correct AND the file is readable
    :param ext: the expected file extension of the file
    :param filep: the filepath of the file
    :return: True if approved, False if not
    """
    approved_exts = ['.csv', '.tsv']
    delim = ''  # namespace scope expansion
    problem = 'none'  # by default, no problems
    for item in approved_exts:
        if ext == item:
            if item is '.csv':
                delim = ','
            elif item is '.tsv':
                delim = '\t'
            else:
                problem = 'some'  # if problem, some problem
    try:
        filep.encode('utf-8').strip()
        pd.read_table(filep, delim, header=0, engine='python')
    except Exception as exc:  # if problem, some problem
        logging.error(exc)
        problem = 'some'
    if problem == 'none':
        return True
    elif problem == 'some':
        return False


def previous_element(_list, index):
    """
    Defines the previous element in a list
    """
    return _list[index-1]


def find_download_links(url, filetype, output_name, in_number=0, download=False, clobber=True):

    """looks around webpages for download links
    :param url: url to be scraped
    :param filetype: filetype to look for in the scraped url
    :param output_name: name of the file to be downloaded
    :param in_number: the number by order of the link in the url
    :param download: if the function should download a file or just return
    :param clobber: if files with the same name should be clobbered
    :return: a dictionary that returns an error as 'error', the outputname as 'filename'
             and the list of links in the url as 'href_list'
    """
    error = 'None'
    p_url = urlparse(url)
    domain = '{urm.scheme}://{urm.netloc}'.format(urm=p_url)  # using outdated urllib1 - this line obtains domain
    dl_name = output_name  # set to the output name by default if no transformations are needed
    link_list = []
    no_tags = ''
    ext_length = len(filetype)
    extension = url[-ext_length:]
    output_path = os.getcwd() + '/data/' + dl_name
    log_message = 'file downloaded succesfully as ' + dl_name + extension + 'from ' + no_tags
    if filetype == output_name[-ext_length:]:   # Format the file name so user input is flexible
        dl_name = output_name[:len(output_name)-ext_length]   # Can include file extension or none
    # First and foremost, check if the user is entering a valid data format
    if extension == filetype:
        link_list.append(url)   # Before anything, check if the url entered IS a dl link
        if download and clobber:  # check if directory exists, check if file is parseable, check allowed file extensions
            urllib.request.urlretrieve(url, output_path)
            if not download_approved(extension, output_path):
                os.remove(output_path)
            else:
                prlog('file downloaded successfully as ' + dl_name)
    else:
        r = urllib.request.urlopen(url)
        soup = BeautifulSoup(r, 'html.parser')
        for link in soup.find_all('a', string=True):   # look through the link tags as strings
            no_tags = link.get('href')
            logging.info(str(no_tags))
            if filetype == str(no_tags)[-ext_length:]:   # Check the three letter file extension
                if 'http://' not in no_tags and 'https://' not in no_tags:   # If no first part of the url, add it
                    no_tags = domain + '/' + no_tags
                link_list.append(no_tags)
        if download and clobber:
            if no_tags is not None:   # Null check
                if len(link_list) == 1:
                    prlog("One link found: " + no_tags)
                    urllib.request.urlretrieve(no_tags, output_path)
                    if not download_approved(extension, output_path):
                        os.remove(output_path)
                    else:
                        prlog(log_message)
                elif len(link_list) > 1:
                    for idx, item in enumerate(link_list):
                        prlog(str(idx) + '. ' + item)
                    if int(in_number) <= len(link_list):
                        no_tags = link_list[int(in_number)]
                        urllib.request.urlretrieve(no_tags, output_path)
                        if not download_approved(extension, output_path):
                            os.remove(output_path)
                        else:
                            prlog(log_message)
                    else:
                        error = 'No link found with that number (' + str(in_number) + ')'
                else:
                    error = 'No links found!'
            elif filetype not in str(no_tags):
                error = 'No file found for that extension (' + filetype + ')'
    prlog(error)
    return {'error': error, 'download_name': dl_name, 'href_list': link_list}



def file_to_htmltable(filepath, delim=','):
    """read a file and generate an html table
    :return: html table script as 'htmltable' variable
    """
    dataframe = pd.read_table(filepath, delim, header=0, engine='python')
    htmltable = dataframe.to_html(bold_rows=True, escape=True)
    return htmltable


def compare(df1, col, title, x_lb, y_lb,
            fedf=get_fe(),
            year_col='Year',
            html='plotlocal.html',
            render=True,
            urlth='patella'
            ):
    """compares a dataframe to the changes in political parties
    :param df1: the first data frame to be compared.
    :param fedf: the second dataframe to be compared, always fe_list.
    :param col: the data column of the first data frame which holds the dates needed to be compared.
    :param title: the title of the plot to be exported.
    :param x_lb: the title of the x axis for the exported plot.
    :param y_lb: the title of the y axis for the exported plot.
    :param year_col: the titel of column of the data frame input to set as the index
    :param html: the html page to be rendered
    :param render: if the function should return js components for a bokeh plot or just the input/resulting dataframes
    :param urlth: the path the url takes to render
    :return: a render_template object which sends a bokeh js/html/css plot to the page as well as a special div.
    """

    data_col = int(col)-1  # subtract 1 from it so the column counting starts from one in the input
    df1.set_index(year_col, drop=True, inplace=True) # if year_col and type(year_col) is "string" else prlog('year col not string')
    ind_list = [parse(str(int(x))).year for x in df1.index.get_level_values(0).values.tolist()]
    office_yr_list = []
    party_office_list = []
    yrs_in_office = 1
    party_yrs_in_office = 1
    term_st_list = fedf.index.get_level_values(0)
    totalfe_list = []
    totalparty_list = []

    # populate a list with tuples of the year and the index where the year changes by looping through the index of the
    # data frame, which is the year column. This spits out a tuple that looks like (years, number of data points)
    # afterwards , the lists interval and years_list are used to store the parts of the tuple individually
    year_and_interval_list = [(i, previous_element(ind_list, i)) for i, year in
                              enumerate(ind_list) if i >= 1 and previous_element(ind_list, i) != ind_list[i]]
    interval = [tpl[0] for tpl in year_and_interval_list]
    years_list = [tpl[1] for tpl in year_and_interval_list]
    total_chg = [(df1.iloc[0, data_col]-df1.iloc[interval[i], data_col]) if i == 0
                 else (df1.iloc[previous_element(interval, i), data_col] - df1.iloc[interval[i], data_col]) if interval[i] < len(ind_list)
                 else (df1.iloc[previous_element(interval, i), data_col] - df1.iloc[(len(ind_list) - previous_element(interval, i)), data_col])
                 for i, item in enumerate(years_list)]

    # A list of variation from year to year, created by dividing the list at index by the previous element,
    # then multiplying by 100 to get a percentage value.
    cov_list = [total_chg[i]/df1.iloc[interval[i], data_col]*100 if previous_element(total_chg, i) != 0
                else total_chg[i] for i, n in enumerate(years_list)]

    # Create lists of all the parties and years over all the us years by finding the distance from from the first
    # presidential term served to the last and appending incremented years in that range to the totalfe totalparty lists
    for idx, item in enumerate(term_st_list):
        if idx != len(term_st_list)-1:
            dist = int(term_st_list[idx+1]) - int(item)
        else:
            dist = datetime.datetime.now().year - int(item) + 1
        for n in range(dist):
            totalfe_list.append(fedf.iloc[idx, 0])
            totalparty_list.append(fedf.iloc[idx, 1])
    # A list of all years there have been presidents
    us_years = [yr+int(term_st_list[0]) for yr in range(int(term_st_list[-1]) - int(term_st_list[0]))]
    # A list of all the parties over every year of the data set
    party_list = [totalparty_list[i] for i, item in enumerate(us_years) for years in years_list if years == item]
    # A list of all the fes over every year of the data set
    fe_list = [totalfe_list[i] for i, item in enumerate(us_years) for years in years_list if years == item]
    # Make the 'years in office' and 'party in office' lists by finding the number of consecutive
    # a party/fe was in office
    # list comprehension possible?
    for idx, item in enumerate(fe_list):
        if idx >= 1 and fe_list[idx - 1] == fe_list[idx]:   # After at least one loop (to avoid oob error) and if the index doesnt change
            yrs_in_office += 1  # increment 1
        else:
            yrs_in_office = 1  # reset to year 1
        office_yr_list.append(yrs_in_office)   # append the year number
    for idx, item in enumerate(party_list):
        if idx >= 1 and party_list[idx - 1] == party_list[idx]:  # After at least one loop (to avoid oob error) and if the index doesnt change
            party_yrs_in_office += 1  # increment 1
        else:
            party_yrs_in_office = 1  # reset to year 1
        party_office_list.append(party_yrs_in_office)   # append the year number

    # Create the dataframe 'plotframe' that will be used to create the bokeh graph
    plotframe = pd.DataFrame({'foo': []})
    plotframe['Total Change'] = total_chg
    plotframe['Percent Change'] = cov_list
    plotframe['Years'] = years_list
    plotframe['Party'] = party_list
    plotframe['First Executive'] = fe_list
    plotframe['Years in Office'] = office_yr_list
    plotframe['Years Party in Office'] = party_office_list
    plotframe.set_index('foo', drop=True, inplace=True)  # remove stock column in a roundabout way
    plotframe.set_index('Years', drop=True, inplace=True)  # remove the year column and set as index
    grouped = plotframe.groupby('First Executive')

    # Bokeh Plot
    p = plotting.figure(title=title,
               x_axis_label=x_lb,
               y_axis_label='Percent Change in ' + y_lb,
               tools="pan,box_zoom,reset,save",
               toolbar_location='below',
               toolbar_sticky=False,
               plot_height= 800,
               plot_width= 800

               )
    # use the plasma palette from bokeh using the number of lines in the figure to create a palette from plasma256
    palette = [palettes.viridis(len(grouped))[i] for i, _ in enumerate(grouped)]
    for idx, (name, data) in enumerate(grouped):
        p.line(x=data['Years in Office'],
               y=data['Percent Change'],
               color=palette[idx],
               legend=name,
               )
    p.background_fill_color = "LightGrey"
    script, div = components(p)  # split the graph into JSON/JS components script and div

    # convert the datframe to html for easy display
    df_htmltable = plotframe.to_html(bold_rows=True, escape=True, classes='dftable')
    if render:
        # If rendering, display the web page
        plotting.show(p)
        return render_template(html, script=script, div=div, table=df_htmltable, var=urlth)

    else:
        # otherwise, return the result dataframe and the input dataframe
        return {'plotframe': plotframe, 'input_frame': df1}



"""
This module is for reading the html tables that list world leaders off of varying webpages.
It is not very dynamic or robust, and has priority for updates.

More tables of FEs from countries around the world will be added so users can compare their data sets to different
world leaders.
"""

import pandas as pd

# A website with an html table to scrape for presidential term data
fe_url = 'http://www.enchantedlearning.com/history/us/pres/list.shtml'
fe_term_column = 2
header_row = 0
unique_text = 'Vice-President'
fe_table = pd.read_html(fe_url, match=unique_text, flavor='bs4', header=header_row,
                        index_col=fe_term_column, parse_dates=True)[0]
fe_names = []
index_list = fe_table.index.get_level_values(0).values.tolist()  # slice the data frame index as a list
year_list = []


for term in index_list:
    term_num = term.split('-')[0]  # get the first year of the term number
    year_list.append(term_num)  # add it to the year list

fe_table['Year'] = year_list
fe_table.set_index('President', drop=False, inplace=True)

#  Some formatting transformations are  required for this table. This loop sets the fe names to only alpha chars
for item in fe_table.index.get_level_values(0).values.tolist():
    fe_name = ''  # reset the name
    for char in item:  # loop through each character to check if it is alpha-nonnumeric, and append it to a string
        if char.isalpha() or char == ' ':
            fe_name = fe_name + char
    fe_names.append(fe_name)  # append the string to the presidential names list
fe_table['President'] = fe_names  # add the name list to the 'President' column of the fe dataframe
fe_table.set_index('Year', drop=True, inplace=True)  # Set the index to the year column


# A simple getter function to return the fe dataframe when needed in other modules
def get_fe():
    return fe_table



from setuptools import setup, find_packages
from codecs import open
from os import path

here = path.abspath(path.dirname(__file__))
with open(path.join(here, 'README.rst'), encoding='utf-8') as f:
    long_description = f.read()

setup(
    name='patella',
    packages=find_packages(),
version='0.1.12',
    author='Salvador Brandi',
    author_email='salbrandi@gmail.com',
    url='https://github.com/salbrandi/patella',
    download_url='https://github.com/salbrandi/patella/archive/0.1.tar.gz',
    py_modules=['htmlparser'],
    description='A webservice for scraping files and plotting data against presidential party variation',
    long_description=long_description,
    install_requires = ['Click',
                        'bokeh',
                        'flask',
                        'bs4',
                        'pandas',
                        'matplotlib',
                        'html5lib',
                        'dateparser'],
    entry_points='''
        [console_scripts]
        patella=patella.click_commands:patella
        ''',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Programming Language :: Python :: 3 :: Only',
    ],
    keywords='plotting data party presidential office term year bokeh'
)



# -*- coding: utf-8 -*-
"""
Controls command line operations

The only particularly relevant command now i:  patella startup <path>

not all commands retain functionality - this will be updated eventually (read: it might not be)

"""

# \/ Third-Party Packages \/
import os
import os.path

import click
import pandas as pd

# \/ Local Packages \/
from . import htmlparser as htmlparser
from . import patellaserver as flaskapp


class filec:
    pass
file1 = filec()
file2 = filec()
file1.df = file2.df = pd.DataFrame({'foo': []})
file1.path = file2.path = ''
file1.name = file2.name = ''


@click.group()
def patella():
    pass


@click.command()
@click.argument('url')
@click.option('--filename', default='datafile', help='specify the name of the local file that will be downloaded to the current directory')
@click.option('--filetype', default='.csv', help='specify the file type the scraper will look for')
def scrape_url(url, filetype, filename):
    parseobj = htmlparser.find_download_links(url, filetype, filename, download=True)
    if type(parseobj) != 'NoneType':
        click.echo('ERROR: ' + parseobj['error'])  # Error reporting


@click.command()
@click.argument('file_one')
@click.option('--delimiters', default=',:,', help='Specify file type delimiters in format <DELIM>:<DELIM2>')
def load_data(file_one, delimiters):
    file1.path = os.getcwd() + '/' + file_one
    if os.path.exists(file1.path):
        file1.name = file_one
        list_delims = delimiters.split(':')
        if len(list_delims) == 2:
            file1.df = pd.read_table(file1.path, list_delims[0], header=0)
            file2.df = htmlparser.get_fe()
            os.environ['LOCAL_FILE_PATH'] = file1.path
            click.echo('file successfully loaded into Dataframes')
    else:
        if not os.path.exists(file1.path):
            click.echo('no files found with the name ' + file_one + ' in path ' + file1.path)



@click.command()
@click.argument('column')
@click.argument('filename')
def change_index(filename, column):
    if filename == file1:
        file1.df.set_index(column)
    else:
        click.echo('no file found with that name')


@click.command()
@click.argument('column_names')
@click.argument('file')
def change_names(file, column_names):
    pass


@click.command()
@click.argument('path')
def startserver(path):
    flaskapp.startserver(path)


@click.command()
@click.argument('file')
@click.argument('col')
@click.option('--title', default=' ', help='specify the plot title')
@click.option('--x_title', default=' ', help='specify the X axis title')
@click.option('--y_title', default=' ', help='specify the Y axis title')
def plot(file, col, title, x_title, y_title):
    file1.path = os.getcwd() + '/data/' + file
    file1.df = pd.read_table(file1.path, ',', header=0)
    htmlparser.compare(file1.df, htmlparser.get_fe(), col, title, x_title, y_title)


# A test cli command
@click.command()
@click.argument('foo')
def testme(foo):
    pass

# add all the subcommands to the patella group
patella.add_command(scrape_url, name='scrape')
patella.add_command(testme, name='test')
patella.add_command(plot)
patella.add_command(load_data, name='load')
patella.add_command(startserver, name='startup')


# -*- coding: utf-8 -*-
#!/usr/bin/env python

"""
This module controls all flask function - from startup to routing urls to the correct html file.
the html files name, functions, and elements,  for reference

index.html: index is the home page. It has 4 links on it - to input.html, to github, and two placeholder routes to input
input.html: input is the home page for patella. it has 2 forms with 2 text fields each, and two submit buttons.
            form 1:
                text field 1: name: url, POSTs the url to be scraped
                text field 2: name: filetype, POSTs the filetype to be looked for
            form 2:
                text field 1: name: filepath, POSTs the local filepath to be plotted
                text field 2: name: datacol, POSTs the column to use for data
options.html: options is the page where the user selects the correct link or table. It has a lsit of radio buttons
              and a text field to specify the file download name, along with a submit button.
              form 1:
                radio field: name:dlink, list of links as radio buttons, POSTs selected
                text field: name: outname, POSTs name of downloaded file
table.html: table is the page which displays the table  generated from the previously downloaded file. it has 2 text
            fields which allows the user to specify the data and date columns
            form 1:
                text field 1: name = datacol, POSTs data column to be used
                text field 2: name = yrcol, POSTs the date column to be used
                text field 3: name = xlab, POSTs the X-axis label to be used
                text field 4: name = ylab, POSTs the Y-Axis label to be used
plotlocal.html: plotlocal is the page which actually displays the plot. it has no forms, just the bokeh js plot
                components along with the table used to generate it

Every html table has a footer with a link which routes back to /input, while /input has a link to /index, the home page.
"""


# Local Imports
from . import htmlparser as htmlparser

# Third Party Imports
from flask import Flask, render_template, request
import pandas as pd
import os
import pathlib

app = Flask(__name__)

if os.path.isdir(os.getcwd().join('data')):
    os.makedirs("data")

app.config['UPLOAD_FOLDER'] = 'data/'
x_axis_label = ''
y_axis_label = ''
title = ''
urlpath = 'patella'


def startserver(path):
    """starts the webservice with specified prefix
    """
    global urlpath
    urlpath = path
    app.run(debug=True, host='0.0.0.0', port=4444)



@app.route('/')
@app.route('/index')
def index():
    """The 'home' page route
    """
    return render_template('index.html', name=urlpath)


@app.route('/<string:var>/input', methods =['POST', 'GET'])
def flask_scrape(var):
    """
    The input route for inputting urls and files
    """
    return render_template('input.html', var=var)



@app.route('/<string:var>/options', methods=['POST', 'GET'])
def options(var):
    """The route for picking download links
    """
    if request.method == 'POST':
        url = request.form['url']
        filetype = request.form['filetype']

        prompt = 'Choose a link to download:'
        button = 'Download and Create Table'
        filename = 'Name of downloaded file'
        try:
            parseobj = htmlparser.find_download_links(url, filetype, 'datafile.csv', download=False)
            result = parseobj['href_list']
            return render_template('options.html',
                                   result=result,
                                   ftype=filetype,
                                   var=var,
                                   prompt=prompt,
                                   button=button,
                                   filename=filename)
        except Exception as exc:
            htmlparser.prlog(exc)
            return render_template('input.html', error=exc, var=var)




@app.route('/<string:var>/table', methods=['POST', 'GET'])
def table(var):
    """The route which allows users to specify options
    """
    if request.method == 'POST':
        dlname = request.form['dlink']
        outname = request.form['outname']
        print(outname)
        htmlparser.find_download_links(dlname[:-1], '.csv', outname, download=True)
        table = htmlparser.file_to_htmltable(os.getcwd() + '/data/' + outname)
        return render_template('table.html', linkname=dlname, table=table, var=var)#['template']


@app.route('/<string:var>/plot', methods=['POST', 'GET'])
def plot_from_df(var):
    """The route to the final plot generated
    """
    error = 'Sorry, an error has occured, and were not sure what is is (>T-T<)'
    if request.method == 'POST':
        result = request.form # store the form results as result
        data_column = result['datacol']  # get the data column from html form
        year_column = result['yearcol'] # get the year column from the html form
        x_axis_label = result['xlab']
        y_axis_label = result['ylab']
        filepath = os.getcwd() + '/data/' + 'datafile.csv'
        if os.path.isfile(filepath):
            try:
                df = pd.read_table(filepath, ',', header=0, engine='python')
                return htmlparser.compare(df,
                                          data_column,
                                          'Percent Change in ' + y_axis_label + ' vs ' + x_axis_label,
                                          x_axis_label,
                                          y_axis_label,
                                          year_col=year_column,
                                          urlth=var)  # Return compare() which returns a render_template() object
            except Exception as exc:
                error = exc
                htmlparser.prlog(error)
                return render_template('input.html', error=error, var=var)
        else:
            htmlparser.prlog(error)
            return render_template('input.html', error=error, var=var)
    

# If run directly, start the server
if __name__ == '__main__':

    startserver(urlpath)
